"""
Class that directly handles the log grid & related maths
"""
import typing

if typing.TYPE_CHECKING:  # for Sphinx
    import numpy as np

    np.zeros(0)

from pyloggrid.Libs.singlethread_numpy import np

np.zeros(0)  # Do not remove - used to force singlethread_numpy to be imported before the other libraries when refactoring

import ctypes as ct
import logging
import math as libmath
import os
import pathlib
from typing import Callable, Iterable, Union

import numpy.ctypeslib as npct
import scipy.optimize

# noinspection PyProtectedMember
from numpy._typing import ArrayLike

import pyloggrid.LogGrid.compute_convolution_kernel as conv_kernel_generator
from pyloggrid.Libs.custom_logger import setup_custom_logger
from pyloggrid.Libs.misc import TimeTracker, bytes2human

logger = setup_custom_logger(__name__, level=logging.INFO)

convolver_c = npct.load_library("convolver_c", pathlib.Path(os.path.abspath(__file__)).parent)


def _setup_convolver_c(convolver_c) -> None:
    """Setup the imported C function. Important to avoid segfaults.
    We keep the convolver as an argument to allow importing this function elsewhere
    """

    npflags = ["C_CONTIGUOUS"]  # Require a C contiguous array in memory
    # noinspection PyTypeChecker
    ui32_1d_type = npct.ndpointer(dtype=np.uint32, ndim=1, flags=npflags)
    # noinspection PyTypeChecker
    complex_1d_type = npct.ndpointer(dtype=complex, ndim=1, flags=npflags)

    args = [ui32_1d_type, ct.c_uint32, complex_1d_type, complex_1d_type, complex_1d_type]
    convolver_c.convolve.argtypes = args
    convolver_c.convolve_omp.argtypes = args

    args = [ui32_1d_type, ct.c_uint32, ct.POINTER(ct.c_void_p), ct.POINTER(ct.c_void_p), ct.c_uint32, ct.c_short, complex_1d_type]
    convolver_c.convolve_list_omp.argtypes = args
    convolver_c.convolve_list_batch_V2.argtypes = args
    convolver_c.convolve_list_batch_V3.argtypes = args
    convolver_c.convolve_list_batch_V4.argtypes = args
    convolver_c.convolve_list_batch_V2_omp.argtypes = args
    convolver_c.convolve_list_batch_V3_omp.argtypes = args
    convolver_c.convolve_list_batch_V4_omp.argtypes = args

    convolver_c.set_omp_threads.argtypes = [ct.c_uint32]


_setup_convolver_c(convolver_c)


class Grid:
    """Main class that handles the log grid"""

    def init_fields(self) -> None:
        """Creates the arrays corresponding to the field names"""
        dict_fields = {field: np.zeros(self.shape, dtype="complex") for field in self.fields_names}
        self.fields = dict_fields

    def get_l_from_params(self) -> float:
        """Sets the grid spacing from the grid's parameters.

        Returns:
            the grid spacing

        See :class:`pyoggrid.LogGrid.Framework.Solver` for more information.

        Reference:
            Campolina & Mailybaev 2020, *Fluid dynamics on logarithmic lattices*
        """
        if self.l_params.get("plastic", False):  # lambda=plastic
            return 1.324717957244746025
        a, b = self.l_params["a"], self.l_params["b"]
        if a is None and b is None:  # lambda=2
            return 2
        # lambda^b-lambda^a=1
        assert int(a) == a and int(b) == b, "a, b should be integers"
        assert b > a, "grid parameter `b` should be greater than `a`"
        assert a > 0, "grid parameter `a` should be >0"
        assert libmath.gcd(a, b) == 1, "grid parameters `a, b` should be irreductible (gcd=1)"
        assert (a != 1 or b != 3) and (a != 4 or b != 5), "(a,b) = (1,3) or (4,5) correspond to plastic number solutions"
        # noinspection PyUnresolvedReferences
        return float(scipy.optimize.fsolve(lambda x: x**b - x**a - 1, np.array(2))[0])

    def __init__(
        self,
        D: int,
        l_params: dict[str, Union[float, bool]],
        N_points: int,
        fields_name: list[str],
        k_min: float = None,
        k0: bool = False,
        n_threads: int = None,
    ):
        """
        Args:
            D: space dimension
            l_params: params that define the grid spacing ``{a, b, plastic: bool}``. For ``l=2``, chose ``a=b=None``. ``plastic`` supercedes all other
            N_points: length of the grid among the smallest axis
            k_min: minimum k on the grid
            fields_name: name of the fields. Used for ordering
            k0: whether the mode ``k=0`` exists
            n_threads: number of threads used in convolutions
        """
        # default val
        if k_min is None:
            k_min = 2 * np.pi
        if n_threads is None:
            try:
                n_threads = int(len(os.sched_getaffinity(0)) * 0.5)
            except AttributeError:
                n_threads = 5

        self.D = D
        self.l_params = l_params
        self.l = self.get_l_from_params()
        logger.debug(f"Chose l={self.l} from params {self.l_params}")
        self.N_points = N_points
        self.k_min = k_min
        self.k0 = k0
        self.ks_1D, self.ks_1D_mid, self.ks, self.ks_modulus = self.generate_points(k_min)
        self.shell_pts = np.zeros((self.ks_1D.size, 2), dtype=object)
        for i, k in enumerate(self.ks_1D):
            pts = (self.ks_modulus == 0) if k == 0 else (k <= self.ks_modulus) & (self.ks_modulus < self.l * k * 0.99)  # 0.99 to avoid overflow
            self.shell_pts[i] = [pts, np.sum(pts)]
        self.k_nonzero = self.ks_modulus > 0
        self.k_alongk0axis = self.ks[0] == 0
        self.maths = Maths(self, n_threads)
        self.physics = Physics(self)
        self.time_tracker: TimeTracker = TimeTracker()

        self.shape = self.ks[0].shape
        self.fields_names = fields_name
        self.fields = None
        self.init_fields()  # init fields with zeros

        self.L = 2 * np.pi / self.k_min
        self.V = self.L**self.D

    def generate_points(self, k_min: float) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Create the grid's wavevector arrays

        Args:
            k_min: the minimum wavevector

        Returns:
            (array of wave vectors along X axis ``[N,]``, logmean of array of wave vectors along X axis [N,], array of wave vectors ``[D, N, (2N, 2N)]``, array of wave vector modulus ``[N, (2N, 2N)]``)
        """
        D = self.D
        N = self.N_points
        l = self.l

        n_k0 = 1 if self.k0 else 0

        points = np.zeros((2 * N + n_k0,))
        for n in range(N):
            points[N + n + n_k0] = l**n
            points[N - 1 - n] = -(l**n)
        points = np.array(points) * k_min

        ks_space = [np.zeros((2 * N + n_k0,) * D)]
        for p in range(len(points)):
            ks_space[0][p] = points[p]
        ks_space.extend(np.swapaxes(ks_space[0], 0, i) for i in range(1, D))
        ks_space = np.array([ks[N:] for ks in ks_space])
        ks_modulus = np.sqrt(np.sum([ks_i**2 for ks_i in ks_space], axis=0))

        ks_1D = points[N:]
        ks_1D_mid = np.zeros_like(ks_1D)
        for i, v in enumerate(ks_1D):
            if i + 1 == ks_1D.size:
                ks_1D_mid[i] = np.exp((np.log(v) + np.log(np.max(ks_modulus))) / 2)
            else:
                ks_1D_mid[i] = 0 if v == 0 else np.exp((np.log(v) + np.log(ks_1D[i + 1])) / 2)

        return ks_1D, ks_1D_mid, ks_space, ks_modulus

    def load_fields(self, fields: dict[str, np.ndarray]) -> "Grid":
        """Load fields from a grid of another dimension. If the new grid is smaller, discard outer fields. If bigger, set new fields to 0.

        Args:
            fields: fields from the previous grid. fields names must correspond for both grids, as well as ``k_min``.
        """
        for field_name, field in self.fields.items():
            N2 = fields[field_name].shape[0]
            n_k0 = 1 if self.k0 else 0
            N = self.N_points + n_k0
            if self.D == 1:
                self.fields[field_name][: min(N2, N)] = fields[field_name][: min(N2, N)]
            elif self.D == 2:
                self.fields[field_name][: min(N2, N), N - min(N2, N) : N + min(N2, N) - n_k0] = fields[field_name][
                    : min(N2, N), N2 - min(N2, N) : N2 + min(N2, N) - n_k0
                ]
            elif self.D == 3:
                self.fields[field_name][: min(N2, N), N - min(N2, N) : N + min(N2, N) - n_k0, N - min(N2, N) : N + min(N2, N) - n_k0] = fields[field_name][
                    : min(N2, N), N2 - min(N2, N) : N2 + min(N2, N) - n_k0, N2 - min(N2, N) : N2 + min(N2, N) - n_k0
                ]
        return self

    def enforce_grid_symmetry(self) -> None:
        """Force the ``f(-k)=f(k).conj`` symmetries along 0 axes for all fields"""
        self.maths.enforce_grid_symmetry_dict(self.fields)

    def field(self, *args) -> list[np.ndarray] | np.ndarray:
        """get fields by name"""
        return [self.fields[i] for i in args] if len(args) > 1 else self.fields[args[0]]

    def to_new_size(self, fields: dict) -> "Grid":
        """return a new grid with the same parameters as this one but a new size

        Args:
            fields: the new fields
        """
        N_points = next(iter(fields.values())).shape[0] - (1 if self.k0 else 0)
        grid = self.to_new_size_empty(N_points)
        grid.fields = fields
        grid.time_tracker = self.time_tracker
        return grid

    def to_new_size_empty(self, N_points: int) -> "Grid":
        """return a new grid with the same parameters as this one but a new size, empty

        Args:
            N_points: the new size
        """
        grid = Grid(
            D=self.D, N_points=N_points, fields_name=self.fields_names, l_params=self.l_params, n_threads=self.maths.n_threads, k_min=self.k_min, k0=self.k0
        )
        grid.time_tracker = self.time_tracker
        return grid


class Maths:
    """Functions to perform maths on log grids

    Attributes:
        convolve_batch (Callable): Convolve a list of convolutions at once. More efficient that calling :func:`convolve` several times. Use as ``convolve_batch([(f0,g0), (f1,g1), ...])``
    """

    cached_conv_kernel = None

    def __init__(self, grid: "Grid", n_threads: int = 0):
        self.grid = grid
        self.n_threads = n_threads
        convolver_c.set_omp_threads(
            n_threads
        )  # In theory we want to minimize the number of calls to this (slow), but we're not supposed to regenerate this object each step so we should be ok
        self.convolve_c = convolver_c.convolve if self.n_threads == 1 else convolver_c.convolve_omp
        self.convolve_c_batch_V = (
            {
                2: convolver_c.convolve_list_batch_V2,
                3: convolver_c.convolve_list_batch_V3,
                4: convolver_c.convolve_list_batch_V4,
            }
            if self.n_threads == 1
            else {
                2: convolver_c.convolve_list_batch_V2_omp,
                3: convolver_c.convolve_list_batch_V3_omp,
                4: convolver_c.convolve_list_batch_V4_omp,
            }
        )
        self.convolve_batch = self._convolve_batch_V if self.n_threads == 1 else self._convolve_batch_list  # list is faster when parallelized (for now)

        self.convolution_kernel = self.generate_convolution_kernel()  # Kernel for convolutions

    def generate_convolution_kernel(self) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Generate the convolution kernel (who interacts with who) for log grid convolutions

        Returns:
            (kernel offsets, kernel signs) | signs are used to determine when to take the complex conjugate (+1 = normal, -1 = conjugate)
        """
        if (
            self.cached_conv_kernel is not None
            and self.cached_conv_kernel["N"] == self.grid.N_points
            and self.cached_conv_kernel["l_params"] == self.grid.l_params
            and self.cached_conv_kernel["k0"] == self.grid.k0
            and self.cached_conv_kernel["D"] == self.grid.D
        ):
            return self.cached_conv_kernel["kernel"]
        a, b = self.grid.l_params["a"], self.grid.l_params["b"]
        if not self.grid.l_params["plastic"] and (a is None and b is None):  # l = 2
            kernel_offsets = [(1, 0), (0, 1), (-1, -1)]
            kernel_signs = [(1, -1), (-1, 1), (1, 1)]
        elif not self.grid.l_params["plastic"]:  # a, b
            kernel_offsets = [(b, a), (a, b), (b - a, -a), (-a, b - a), (-b, a - b), (a - b, -b)]

            kernel_signs = [(1, -1), (-1, 1), (1, -1), (-1, 1), (1, 1), (1, 1)]

        else:  # plastic
            kernel_offsets = [(3, 1), (1, 3), (2, -1), (-1, 2), (-3, -2), (-2, -3), (5, 4), (4, 5), (1, -4), (-4, 1), (-5, -1), (-1, -5)]

            kernel_signs = [(1, -1), (-1, 1), (1, -1), (-1, 1), (1, 1), (1, 1), (1, -1), (-1, 1), (1, -1), (-1, 1), (1, 1), (1, 1)]

        # k=0
        i_list = list(range(-self.grid.ks[0].shape[0] + 1, self.grid.ks[0].shape[0]))
        i_list.remove(0)
        kernel_offsets_k0_ax0 = np.array([(0, 0)] + [(i, i) for i in i_list] * 2)
        kernel_signs_k0_ax0 = np.array(
            [(1, 1)] + [(1, -1) for _ in range((kernel_offsets_k0_ax0.shape[0] - 1) // 2)] + [(-1, 1) for _ in range((kernel_offsets_k0_ax0.shape[0] - 1) // 2)]
        )
        if self.grid.D > 1:
            i_list = list(range(-self.grid.ks[0].shape[1] + 1, self.grid.ks[0].shape[1]))
            i_list.remove(0)
            kernel_offsets_k0_ax1 = np.array([(0, 0)] + [(i, i) for i in i_list] * 2)
            kernel_signs_k0_ax1 = np.array(
                [(1, 1)]
                + [(1, -1) for _ in range((kernel_offsets_k0_ax1.shape[0] - 1) // 2)]
                + [(-1, 1) for _ in range((kernel_offsets_k0_ax1.shape[0] - 1) // 2)]
            )
        else:
            kernel_offsets_k0_ax1, kernel_signs_k0_ax1 = 1, 1
        if self.grid.k0:
            kernel_offsets += [(0, 0), (0, 0)]
            kernel_signs += [(1, 0), (0, 1)]
        # convert for Cython
        kernel_offsets, kernel_signs, kernel_offsets_k0_ax0, kernel_signs_k0_ax0, kernel_offsets_k0_ax1, kernel_signs_k0_ax1 = (
            np.array(kernel_offsets, dtype=np.short),
            np.array(kernel_signs, dtype=np.short),
            np.array(kernel_offsets_k0_ax0, dtype=np.short),
            np.array(kernel_signs_k0_ax0, dtype=np.short),
            np.array(kernel_offsets_k0_ax1, dtype=np.short),
            np.array(kernel_signs_k0_ax1, dtype=np.short),
        )

        if self.grid.D == 1:
            # noinspection PyTypeChecker
            kernel = conv_kernel_generator.compute_interaction_kernel_1D(
                kernel_offsets, kernel_signs, kernel_offsets_k0_ax0, kernel_signs_k0_ax0, self.grid.k0, self.grid.ks[0].shape[0]
            )
        elif self.grid.D == 2:
            # noinspection PyTypeChecker
            kernel = conv_kernel_generator.compute_interaction_kernel_2D(
                kernel_offsets,
                kernel_signs,
                kernel_offsets_k0_ax0,
                kernel_signs_k0_ax0,
                kernel_offsets_k0_ax1,
                kernel_signs_k0_ax1,
                self.grid.k0,
                self.grid.ks[0].shape[0],
                self.grid.ks[0].shape[1],
            )
        elif self.grid.D == 3:
            # noinspection PyTypeChecker
            kernel = conv_kernel_generator.compute_interaction_kernel_3D(
                kernel_offsets,
                kernel_signs,
                kernel_offsets_k0_ax0,
                kernel_signs_k0_ax0,
                kernel_offsets_k0_ax1,
                kernel_signs_k0_ax1,
                self.grid.k0,
                self.grid.ks[0].shape[0],
                self.grid.ks[0].shape[1],
                self.grid.ks[0].shape[2],
            )
        else:
            raise ValueError
        # noinspection PyUnresolvedReferences
        logger.info(f"Convolution coeffs generated - approx {kernel.shape[0]} triads, kernel size approx {bytes2human(kernel.size * kernel.itemsize)}")
        Maths.cached_conv_kernel = {"N": self.grid.N_points, "l_params": self.grid.l_params, "D": self.grid.D, "k0": self.grid.k0, "kernel": kernel}
        # noinspection PyTypeChecker
        return kernel

    # actual operations below
    @property
    def dx(self) -> np.ndarray:
        """x-derivative"""
        return 1j * self.grid.ks[0]

    @property
    def dy(self) -> np.ndarray:
        """y-derivative"""
        return 1j * self.grid.ks[1]

    @property
    def dz(self) -> np.ndarray:
        """z-derivative"""
        return 1j * self.grid.ks[2]

    @property
    def d2x(self) -> np.ndarray:
        """2nd x-derivative"""
        return self.dx * self.dx

    @property
    def d2y(self) -> np.ndarray:
        """2nd y-derivative"""
        return self.dy * self.dy

    @property
    def d2z(self) -> np.ndarray:
        """2nd z-derivative"""
        return self.dz * self.dz

    def convolve(self, f: np.ndarray, g: np.ndarray) -> np.ndarray:
        """Convolve the two arrays.

        Returns:
            the convolved array
        """
        with self.grid.time_tracker("convolution"):
            f_flat = f.flatten()
            res = np.zeros_like(f_flat)
            # noinspection PyUnresolvedReferences
            with self.grid.time_tracker("convolution_C"):
                # noinspection PyUnresolvedReferences
                self.convolve_c(self.convolution_kernel, self.convolution_kernel.size, f_flat, g.flatten(), res)
            res = res.reshape(f.shape)

        return res

    def _convolve_batch_V_inner(self, fgs: list[tuple[np.ndarray, np.ndarray]], V: int) -> np.ndarray:
        """Inner function for <_convolve_batch_V>. Convolves in batch V couples, calling the corresponding ``convolve_list_batch_V`` function in C."""
        if V == 1:
            f, g = fgs[0][0], fgs[0][1]
            f_flat = f.flatten()
            res = np.zeros_like(f_flat)
            with self.grid.time_tracker("convolution_C"):
                # noinspection PyUnresolvedReferences
                self.convolve_c(self.convolution_kernel, self.convolution_kernel.size, f_flat, g.flatten(), res)
            return res
        fsize = fgs[0][0].size
        N_batch = len(fgs)
        res = np.zeros((fsize * N_batch,), dtype=fgs[0][0].dtype)

        f_flats, g_flats = [], []
        for f, g in fgs:
            f_flats.append(f.flatten())
            g_flats.append(g.flatten())

        arr_f, arr_g = (ct.c_void_p * N_batch)(), (ct.c_void_p * N_batch)()  # create array of pointers
        for i in range(N_batch):
            arr_f[i], arr_g[i] = f_flats[i].ctypes.data_as(ct.c_void_p), g_flats[i].ctypes.data_as(ct.c_void_p)

        with self.grid.time_tracker("convolution_C"):
            # noinspection PyUnresolvedReferences
            self.convolve_c_batch_V[V](self.convolution_kernel, self.convolution_kernel.size, arr_f, arr_g, fsize, N_batch, res)

        return res

    def _convolve_batch_V(self, fgs: list[tuple[np.ndarray, np.ndarray]]) -> np.ndarray:
        """from couples ``((f0, g0), (f1, g1), ...)``, computes their convolution

        Returns:
            the convolved arrays ``f0*g0``, ``f1*g1``, etc.
        """
        with self.grid.time_tracker("convolution"):
            max_V = 4  # relevant for most processors. This is not threading ! This is how many convolutions we compute "at the same time" on one thread.
            N_batch = len(fgs)
            N_V, diff_V = N_batch // max_V, N_batch % max_V
            res_maxV, res_diffV = np.array([], dtype=fgs[0][0].dtype), np.array([], dtype=fgs[0][0].dtype)
            if N_V > 0:
                res_maxV = self._convolve_batch_V_inner(fgs[: N_V * max_V], max_V)
            if diff_V > 0:
                res_diffV = self._convolve_batch_V_inner(fgs[-diff_V:], diff_V)

            res = np.concatenate((res_maxV, res_diffV)).reshape((len(fgs),) + fgs[0][0].shape)
        return res

    def _convolve_batch_list_slower(self, fgs: list[tuple[np.ndarray, np.ndarray]]) -> np.ndarray:
        """from couples ``((f0, g0), (f1, g1), ...)``, computes their convolution using batched OMP-supported convolutions

        Returns:
            the convolved arrays ``f0*g0``, ``f1*g1``, etc.
        """
        with self.grid.time_tracker("convolution"):
            fsize = fgs[0][0].size
            N_batch = len(fgs)
            res = np.zeros((fsize * N_batch,), dtype=fgs[0][0].dtype)

            f_flats, g_flats = [], []
            for f, g in fgs:
                f_flats.append(f.flatten())
                g_flats.append(g.flatten())

            arr_f, arr_g = (ct.c_void_p * N_batch)(), (ct.c_void_p * N_batch)()  # create array of pointers
            for i in range(N_batch):
                arr_f[i], arr_g[i] = f_flats[i].ctypes.data_as(ct.c_void_p), g_flats[i].ctypes.data_as(ct.c_void_p)

            with self.grid.time_tracker("convolution_C"):
                # noinspection PyUnresolvedReferences
                convolver_c.convolve_list_omp(self.convolution_kernel, self.convolution_kernel.size, arr_f, arr_g, fsize, N_batch, res)

            res = res.reshape((len(fgs),) + fgs[0][0].shape)
        return res

    def _convolve_batch_list(self, fgs: list[tuple[np.ndarray, np.ndarray]]) -> Iterable[np.ndarray]:
        """from couples ``((f0, g0), (f1, g1), ...)``, computes their convolution

        Returns:
            the convolved arrays ``f0*g0``, ``f1*g1``, etc.
        """
        return [self.convolve(f, g) for (f, g) in fgs]

    @staticmethod
    def cross2D(a: ArrayLike, b: ArrayLike) -> np.ndarray:
        """2D cross product"""
        ax, ay = np.array(a)
        bx, by = b
        return ax * by - ay * bx

    @staticmethod
    def cross3D(a: ArrayLike, b: ArrayLike) -> np.ndarray:
        """3D cross product"""
        ax, ay, az = np.array(a)
        bx, by, bz = b
        return np.array([ay * bz - az * by, bx * az - bz * ax, ax * by - ay * bx])

    def rot2D(self, a: ArrayLike) -> np.ndarray:
        """2D rotational"""
        return 1j * self.cross2D(self.grid.ks, a)

    def rot3D(self, a: ArrayLike) -> np.ndarray:
        """3D rotational"""
        return 1j * self.cross3D(self.grid.ks, a)

    @property
    def rot2D_inv(self) -> np.ndarray:
        """Inverse rotational for 2D fields, *assuming div = 0*"""
        return np.array([-self.dy * self.laplacian_inv, self.dx * self.laplacian_inv])

    def rot3D_inv(self, a: ArrayLike) -> np.ndarray:
        """Inverse rotational for 3D fields, *assuming div = 0*"""
        return -self.laplacian_inv * self.rot3D(a)

    @property
    def laplacian(self) -> np.ndarray:
        """laplacian"""
        return -self.grid.ks_modulus**2

    @staticmethod
    def inv(f: ArrayLike) -> np.ndarray:
        """Inverse of a linear operator.

        it's assumed that where the operator is zero, the function to invert will also be zero
        """
        f = np.array(f)
        nz = f != 0
        f[nz] = 1 / f[nz]
        return f

    @property
    def laplacian_inv(self) -> np.ndarray:
        """inverse laplacian"""
        return self.inv(self.laplacian)

    def div3D(self, a: ArrayLike) -> np.ndarray:
        """3D divergence"""
        ax, ay, az = a
        return self.dx * ax + self.dy * ay + self.dz * az

    def div2D(self, a: ArrayLike) -> np.ndarray:
        """2D divergence"""
        ax, ay = a
        return self.dx * ax + self.dy * ay

    def inner_product(self, f: ArrayLike, g: ArrayLike) -> float:
        """Log grid inner product"""
        mask = np.ones_like(g)
        f = np.array(f)
        if self.grid.k0:
            if len(f.shape) == len(self.grid.ks[0].shape) + 1:
                mask[:, self.grid.k_alongk0axis] = 1 / 2
                # TODO pourquoi 2*1/2 et pas 1* sur la moitié des points ? Et on divise pas à tort le point (0,0) par 2 par hasard ?
            else:
                mask[self.grid.k_alongk0axis] = 1 / 2
        return 2 * np.sum(f * np.conj(g * mask)).real

    def self_inner_product(self, f: ArrayLike) -> float:
        """inner product applied on oneself"""
        return self.inner_product(f, f)  # noqa

    def P_projector(self, A: ArrayLike) -> np.ndarray:
        """Turns ``dUdt = A - gradP`` into ``dUdt = A'``

        Args:
            A: ``dUdt`` without the pressure term

        Returns:
            ``A'``
        """
        ks, ks_modulus = self.grid.ks, self.grid.ks_modulus
        Calc = np.zeros_like(A, dtype="complex")
        if len(Calc.shape) == len(self.grid.ks[0].shape) + 1:
            Calc[:, self.grid.k_nonzero] = (ks * np.sum(A * ks, axis=0))[:, self.grid.k_nonzero] / ks_modulus[self.grid.k_nonzero] ** 2
        else:
            Calc[self.grid.k_nonzero] = (ks * np.sum(A * ks, axis=0))[self.grid.k_nonzero] / ks_modulus[self.grid.k_nonzero] ** 2

        return A - Calc

    def enforce_grid_symmetry_arr(self, arr: ArrayLike) -> np.ndarray:
        """Force the ``f(-k)=f(k).conj`` symmetries along ``k=0`` axes. Modifies the array in-place"""
        arr = np.array(arr)
        grid = self.grid
        D = grid.D
        arr[grid.ks_modulus == 0] = arr[grid.ks_modulus == 0].real  # enfore symmetry
        if D > 1:
            arr[(grid.ks[0] == 0) & (grid.ks[1] < 0)] = arr[(grid.ks[0] == 0) & (grid.ks[1] > 0)][::-1].conjugate()
        if D > 2:
            arr[(grid.ks[0] == 0) & (grid.ks[1] == 0) & (grid.ks[2] < 0)] = arr[(grid.ks[0] == 0) & (grid.ks[1] == 0) & (grid.ks[2] > 0)][::-1].conjugate()

        return arr

    def enforce_grid_symmetry_dict(self, fields: dict) -> dict:
        """forces symmetry for all fields of dict, in-place"""
        if self.grid.k0:  # only enforce if we have k0 mode to avoid slowing down
            for field_name, field in fields.items():
                fields[field_name] = self.enforce_grid_symmetry_arr(field)

        return fields


class Physics:
    """Computes a few classical physical properties"""

    def __init__(self, grid: "Grid"):
        self.grid = grid

    def enstrophy(self) -> float:
        """enstrophy Ω=ω²/2"""
        if self.grid.D != 3:
            raise NotImplementedError("1D/2D isn't coded yet")

        if "wx" in self.grid.fields:
            wx = self.grid.fields["wx"]
            wy = self.grid.fields["wy"]
            wz = self.grid.fields["wz"]
            w = [wx, wy, wz]
        else:
            ux = self.grid.fields["ux"]
            uy = self.grid.fields["uy"]
            uz = self.grid.fields["uz"]
            w = self.grid.maths.rot3D([ux, uy, uz])
        Omega = self.grid.maths.self_inner_product(w) / 2
        return Omega.real

    def energy(self) -> float:
        """kinetic energy E=u²/2"""
        match self.grid.D:
            case 3:
                u = self.grid.maths.rot3D_inv(self.grid.field("wx", "wy", "wz")) if "wx" in self.grid.fields else self.grid.field("ux", "uy", "uz")
            case 2:
                if "w" in self.grid.fields:
                    w = self.grid.fields["w"]
                    u = self.grid.maths.rot2D_inv * w
                else:
                    ux = self.grid.fields["ux"]
                    uy = self.grid.fields["uy"]
                    u = [ux, uy]
            case 1:
                u = self.grid.fields["u"]
            case _:
                raise ValueError

        E = self.grid.maths.self_inner_product(u) / 2
        return E.real

    def helicity(self) -> complex:
        """helicity H=u*ω"""
        if self.grid.D != 3:
            raise NotImplementedError("1D/2D isn't coded yet")

        wx = self.grid.fields["wx"]
        wy = self.grid.fields["wy"]
        wz = self.grid.fields["wz"]
        w = np.array([wx, wy, wz])

        u = self.grid.maths.rot3D_inv(w)
        H = self.grid.maths.inner_product(u, w)
        return H.real

    def spectrum(self, fun: Callable[[dict[str, np.ndarray], np.ndarray], np.ndarray]) -> np.ndarray:
        """Spectrum for the quantity calculated by the callback
        Callback takes two args: ``fields`` and ``ks``. The first one is ``grid.fields``. The second one is a grid-shaped bool array corresponding to concerned ``ks``.
        The callback should return a summable np.ndarray ``[observable(k) for k in ks]``.
        Ex kinetic energy 2D: ``fun(fields, ks) = ux[k]*conj(ux[k]) + uy[k]*conj(uy[k])``

        Returns:
            np.ndarray of the spectrum along each point of
            grid.ks_1D
        TODO update utest
        """
        return self.compute_by_shell(fun, normalize=True)

    def cumulative_k(self, fun: Callable[[dict[str, np.ndarray], np.ndarray], np.ndarray]) -> np.ndarray:
        """Compute the cumulative quantity between wave number 0 and k, defined by the sum over \\|k'| < k of that quantity
        TODO add utest
        """
        return np.cumsum(self.compute_by_shell(fun))

    def compute_by_shell(self, fun: Callable[[dict[str, np.ndarray], np.ndarray], np.ndarray], normalize: bool = False) -> np.ndarray:
        """Structure function for the quantity calculated by the callback
        Callback takes two args: ``fields`` and ``ks``. The first one is ``grid.fields``. The second one is a grid-shaped bool array corresponding to concerned ``ks``.
        The callback should return a summable np.ndarray ``[observable(k) for k in ks]``.
        Ex kinetic energy 2D: ``fun(fields, ks) = ux[k]*conj(ux[k]) + uy[k]*conj(uy[k])``

        Args:
            fun: the function to compute
            normalize: if True, normalize the result on each shell

        Returns:
            np.ndarray of the structure function along each point of
            grid.ks_1D
        TODO: add utest
        """
        grid = self.grid
        result_by_shell = np.zeros((grid.N_points + (1 if grid.k0 else 0),))

        for i, k in enumerate(grid.ks_1D):
            pts, N_k = grid.shell_pts[i]
            if N_k == 0:
                continue
            res = np.sum(fun(grid.fields, pts))
            if normalize:
                res /= (grid.l - 1) * k * N_k if k != 0 else 1
            assert res.imag == 0, "compute_by_shell: got imaginary results"
            result_by_shell[i] += res

        return result_by_shell



"""
Ensures that ``numpy`` is single-threaded. Does not disrupt ``Cython`` threading.

Warning:
    Must be imported top-level, before any other library that uses ``numpy`` !!

Example:
    Replace
    ::

        import numpy as np

    by

    ::

        from pyloggrid.Libs.singlethread_numpy import np
"""
import os
import sys

from pyloggrid.Libs.custom_logger import setup_custom_logger

logger = setup_custom_logger(__name__)

if os.name == "posix":
    import resource  # Increase max recursion depth

    try:
        resource.setrlimit(resource.RLIMIT_STACK, (2**29, -1))
    except ValueError:
        logger.warning("Could not set RLIMIT_STACK")
sys.setrecursionlimit(int(1e6))

os.environ["OPENBLAS_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"

# noinspection PyUnresolvedReferences
import numpy as np_

np = np_  # Otherwise sphinx-autoapi doesn't detect it
np.array(0)  # to avoid F401 in flake8 lint



"""Data processing and visualisation"""

import logging
import os
import typing
from typing import Any, Callable, NotRequired, Optional, TypedDict, TypeVar

if typing.TYPE_CHECKING:  # for Sphinx
    import numpy as np

    np.zeros(0)

from pyloggrid.Libs.singlethread_numpy import np

np.zeros(0)  # Do not remove - used to force singlethread_numpy to be imported before the other libraries when refactoring

import joblib

from pyloggrid.Libs.custom_logger import setup_custom_logger
from pyloggrid.Libs.IOLib import add_drawables, load_step, read_json_settings
from pyloggrid.Libs.misc import TimeTracker
from pyloggrid.LogGrid.Grid import Grid

logger = setup_custom_logger(__name__, level=logging.INFO)


class DataExplorer:
    """
    Data processing and visualisation

    Args:
        data_path: simulation's folder
    """

    def __init__(self, data_path: str):
        self.path = data_path
        assert os.path.isdir(data_path), f"The path {data_path} could not be found"

        self.time_tracker = TimeTracker()
        self.time_tracker.start_timer("total")

        with self.time_tracker("disk"):
            settings = read_json_settings(data_path)

        self.D = settings["D"]
        self.l_params = settings["l_params"]
        self.simu_params = settings["simu_params"]
        self.fields_name = settings["fields_name"]
        self.N_steps = settings["N_steps"]

        logger.info(f"Expecting {self.N_steps} steps to load")

    def getgrid(self, N_pts: int, k_min: float, fields: dict[str, np.ndarray] = None, k0: bool = False) -> Grid:
        """Creates a grid populated with a step's fields.

        Args:
            N_pts: grid size
            k_min: grid's min k
            fields: grid's fields
            k0: whether grid has k=0

        Returns:
            the populated grid
        """
        grid = Grid(D=self.D, l_params=self.l_params, N_points=N_pts, fields_name=self.fields_name, k_min=k_min, k0=k0)
        if fields is not None:
            grid.load_fields(fields)
        return grid

    # noinspection PyMissingOrEmptyDocstring
    class DrawFuncDict(TypedDict):
        get: Callable[["DataExplorer"], dict] | Callable[[Grid, float, dict], dict]
        plot: Callable[[Callable], None]
        perframe: NotRequired[bool]

    def display(
        self,
        draw_funcs: dict[str, DrawFuncDict],
        N_points: Optional[int] = None,
        loadfromsave: bool = False,
        N_min: int = 1,
        N_max: Optional[int] = None,
        n_jobs: int = 1,
    ) -> None:
        """Compute and display all the requested drawables.

        Args:
            draw_funcs: functions to draw, as a dict ``{name -> {"get"->fun, "plot"->fun(drawables, ts), "perframe"->bool}}``.
                If ``perframe``, the getter is per step and has signature ``fun(grid, t, simu_params)``. Otherwise, it is called once with signature ``fun(dataExplorer)``
                If ``perframe`` is omitted, default to ``True``
            N_points: max number of points to evaluate. If ``None``, equal to the max available points
            N_min: The first step evaluated
            N_max: The last step evaluated
            loadfromsave: data is not calculated but directly loaded from already computed drawables in ``drawables.npy``. ``N_min, N_max, N_points`` are not taken into account.
            n_jobs: the number of parrallel threads to use. Use ``-1`` for unlimited (not recommended)
        """
        drawables = {}  # name -> data to draw
        if loadfromsave:  # check if previous data exists, and load if possible
            with self.time_tracker("disk"):
                if os.path.isfile(f"{self.path}/drawables.npy"):
                    logger.info(f"Loading data from {self.path}/drawables.npy")
                    # noinspection PyTypeChecker
                    dict_draw: dict = np.load(f"{self.path}/drawables.npy", allow_pickle=True).item()
                    drawables = dict_draw["drawables"]
                    nameNotInDrawables = []
                    for name, draw_func in draw_funcs.items():
                        if name not in drawables:
                            nameNotInDrawables.append(name)

                    if nameNotInDrawables:
                        logger.warning(f"No data found for {nameNotInDrawables}, recomputing")
                        loadfromsave = False
                else:
                    logger.warning(f"{self.path}/drawables.npy doesn't exist, setting loadfromsave=False")
                    loadfromsave = False

        if not loadfromsave:
            draw_funcs_filtered = {k: v for k, v in draw_funcs.items() if k not in drawables}

            logger.info("Warning: this assumes a divergence-free flow field")

            if N_max is None or N_max > self.N_steps:
                N_max = self.N_steps
            N_points = N_max - N_min + 1 if N_points is None else min(N_points, N_max - N_min + 1)

            draw_perframe = []
            draw_notperframe = []
            for name, draw_func in draw_funcs_filtered.items():
                if ("perframe" not in draw_func) or draw_func["perframe"]:
                    draw_perframe.append([name, draw_func])
                    drawables[name] = []
                else:
                    draw_notperframe.append([name, draw_func])

            def load_time_dependant_drawables(step: int) -> tuple[dict[str, Any], TimeTracker]:
                """
                Load all the time-dependant variables for the given step
                :param step: step number
                :return: dict drawable_name -> drawable result
                """
                logger.info(f"Loading step {step}/{self.N_steps} ({np.round(100 * step / self.N_steps, 2)}%)")
                drawables_perstep = {}
                self.time_tracker = TimeTracker()  # reset and return because parallelization loses the data
                self.time_tracker.start_timer("total")

                # load step data
                with self.time_tracker("disk"):
                    item: dict = self.load_step(step, grid=False)
                t, N_pts, k_min, fields, k0 = item["t"], item["N_points"], item["k_min"], item["fields"], item["k0"]

                # create corresponding grid
                grid = self.getgrid(N_pts, k_min, fields, k0)
                # compute each drawable
                with self.time_tracker("drawables_perstep"):
                    for name, draw_func in draw_perframe:
                        drawables_perstep[name] = draw_func["get"](grid, t, self.simu_params)
                        drawables_perstep[name]["t"] = t

                self.time_tracker.end_timer("total")
                return drawables_perstep, self.time_tracker

            # compute all the time-dependant quantities
            if draw_perframe:
                self.time_tracker.end_timer("total")  # stop counting during parallel step to avoid double counting
                res_parallel = joblib.Parallel(n_jobs=n_jobs)(
                    joblib.delayed(load_time_dependant_drawables)(step) for step in np.round(np.linspace(N_min, N_max, N_points)).astype(int)
                )
                self.time_tracker.start_timer("total")
                for step, (drawable_step, timer_step) in enumerate(res_parallel):
                    # append each data part to their data holder
                    for k, v in timer_step.elapsed_time.items():
                        if k not in self.time_tracker.elapsed_time:
                            self.time_tracker.elapsed_time[k] = 0
                        self.time_tracker.elapsed_time[k] += v
                    for name, draw_func in draw_perframe:
                        drawables[name].append(drawable_step[name])

                # re-aggregate the drawables by label:
                # original: [{"a": 0, "b": 5}, {"a": 2, "b": -1}]
                # new: {"a": [0, 2], "b": [5, -1]}
                for name, _ in draw_perframe:
                    drawables_bk = np.array(drawables[name])
                    drawables[name] = {}
                    for key in drawables_bk[0].keys():
                        drawables[name][key] = []
                    for i in range(N_points):
                        for k, v in drawables_bk[i].items():
                            drawables[name][k].append(v)
                    for key in drawables_bk[0].keys():
                        try:
                            drawables[name][key] = np.array(drawables[name][key])
                        except ValueError:
                            drawables[name][key] = np.array(drawables[name][key], dtype="object")

            # compute global quantities
            with self.time_tracker("drawables_global"):
                for name, draw_func in draw_notperframe:
                    logger.info(f"Calculating {name}")
                    drawables[name] = draw_func["get"](self)

            # Save all data as npy
            with self.time_tracker("disk"):
                add_drawables(self.path, drawables)

        # Plot
        self.time_tracker.end_timer("total")
        logger.info(f"Done ! Now plotting - elapsed: {self.time_tracker}")
        for name, draw_func in draw_funcs.items():
            # noinspection PyUnboundLocalVariable
            draw_func["plot"](lambda *args: ((drawables[name][i] for i in args) if len(args) > 1 else drawables[name][args[0]]))

    def load_step(self, step: int = None, ts: np.ndarray = None, t: float = None, grid: bool = True) -> dict | tuple[dict, Grid]:
        """
        Load a saved step, either by step number, or by simulation time. If time is provided, it overrides the step.

        Args:
            step: step to load
            ts: array of simulation times
            t: time of the step to use. Must be provided with ``ts``. Overrides ``step``.
            grid: if ``True``, create a grid from loaded data and return it

        Returns:
            step data ``{fields, t, N_points, k_min, elapsed_time}`` if not ``grid``, else a tuple ``step_data, grid``
        """
        if t is not None:
            assert ts is not None, "`ts` must be provided with t"
            step_resc = np.argmax(ts >= t)
            step = int(step_resc * self.N_steps / ts.size)
            logger.info(f"Loading timed step: {step}, t={t}")
        elif step is None:
            step = self.N_steps

        data = load_step(fields_names=self.fields_name, path=self.path, step=step)
        if not grid:
            return data
        N_points, fields, k_min = data["N_points"], data["fields"], data["k_min"]
        return data, self.getgrid(N_points, k_min, fields)


T = TypeVar("T", bound=str)
PlotFun = Callable[..., tuple[np.ndarray, ...]]



"""Misc small libraries that help with various tasks, too small to fit in their own library file"""
from __future__ import annotations

import time

from pyloggrid.Libs.singlethread_numpy import np


def composed_decs(*decs):
    """Compose several decorators into one.

    Example:
        ::

            @composed(dec_1, dec_2)
            def f(...)

        is equivalent to

        ::

            @dec_1
            @dec_2
            def f(...)
    """

    # noinspection PyMissingOrEmptyDocstring
    def deco(f):
        for dec in reversed(decs):
            f = dec(f)
        return f

    return deco


def bytes2human(n, format_="%(value).1f %(symbol)s", symbols="customary"):
    """Convert n bytes into a human readable string based on format.
    symbols can be either "customary", "customary_ext", "iec" or "iec_ext",
    see: https://goo.gl/kTQMs

    Example:
        ::

            > bytes2human(0)
            '0.0 B'
            > bytes2human(0.9)
            '0.0 B'
            > bytes2human(1)
            '1.0 B'
            > bytes2human(1.9)
            '1.0 B'
            > bytes2human(1024)
            '1.0 K'
            > bytes2human(1048576)
            '1.0 M'
            > bytes2human(1099511627776127398123789121)
            '909.5 Y'

            > bytes2human(9856, symbols="customary")
            '9.6 K'
            > bytes2human(9856, symbols="customary_ext")
            '9.6 kilo'
            > bytes2human(9856, symbols="iec")
            '9.6 Ki'
            > bytes2human(9856, symbols="iec_ext")
            '9.6 kibi'

            > bytes2human(10000, "%(value).1f %(symbol)s/sec")
            '9.8 K/sec'

            > # precision can be adjusted by playing with %f operator
            > bytes2human(10000, format_="%(value).5f %(symbol)s")
            '9.76562 K'
    """
    SYMBOLS = {
        "customary": ("B", "K", "M", "G", "T", "P", "E", "Z", "Y"),
        "customary_ext": ("byte", "kilo", "mega", "giga", "tera", "peta", "exa", "zetta", "iotta"),
        "iec": ("Bi", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi"),
        "iec_ext": ("byte", "kibi", "mebi", "gibi", "tebi", "pebi", "exbi", "zebi", "yobi"),
    }

    n = int(n)
    assert n >= 0
    symbols = SYMBOLS[symbols]
    prefix = {s: 1 << (i + 1) * 10 for i, s in enumerate(symbols[1:])}
    for symbol in reversed(symbols[1:]):
        if n >= prefix[symbol]:
            value = float(n) / prefix[symbol]
            return format_ % locals()
    return format_ % dict(symbol=symbols[0], value=n)


class TimeTracker:
    """Keep track of elapsed time. Supports nesting.

    How to use:

    * to start a timer, use ``timer.start_timer(key)`` where ``key`` is a unique identifier
    * to stop a timer and add the elapsed time to the total for the key, use ``timer.stop_timer(key)``
    * to stop and immediately restart (used for storing the current value) a timer use ``timer.tick_timer(key)``

    Wherever possible, prefer wrapping you code in a with block:
    ::

        with timer(key):
            ...

    """

    def __init__(self, initdict: dict = None):
        self.elapsed_time = initdict or {}  # total elapsed time
        self.timer_start = {}  # start of each tracker
        self.current_key = []  # list of current keys to enable nesting with blocks

    def __call__(self, key: str):
        self.current_key.append(key)
        return self

    def __enter__(self):
        self.start_timer(self.current_key[-1])

    def __exit__(self, *args):
        self.end_timer(self.current_key.pop())

    def get(self, key: str) -> float:
        """Get elapsed time for ``key``."""
        return self.elapsed_time.get(key, 0)

    def start_timer(self, key: str) -> None:
        """Start timer for ``key``"""
        self.timer_start[key] = time.perf_counter()

    def end_timer(self, key: str) -> float:
        """Stop timer for ``key>``, store and return total elapsed time."""
        elapsed = time.perf_counter() - self.timer_start[key]
        if key not in self.elapsed_time:
            self.elapsed_time[key] = 0
        self.elapsed_time[key] += elapsed
        return self.elapsed_time[key]

    def tick_timer(self, key: str) -> None:
        """Store elapsed time and reset timer."""
        self.end_timer(key)
        self.start_timer(key)

    def __repr__(self):
        def format_time(seconds: float) -> str:
            """human-readable"""
            # noinspection PyUnresolvedReferences
            delta_dict = {
                "d": seconds // (3600 * 24),
                "h": seconds // 3600,
                "m": seconds // 60,
                "s": seconds // 1,
                "ms": np.round(1000 * (seconds % 1), 3).astype(int),
            }
            res = " ".join(f"{int(v)}{k}" for k, v in delta_dict.items() if v)
            return res or "0ms"

        return "{" + ", ".join([f"{k}: {format_time(v)}" for k, v in sorted(self.elapsed_time.items(), key=lambda item: -item[1])]) + "}"


def rightpad_array_2D(a: np.ndarray, Nx: int, Ny: int) -> np.ndarray:
    """Pad a 2D+ array with zeros.

    The padding is done at the large coordinate edge (it is not centered).

    Args:
        a: array to pad
        Nx: amount of pixels to pad in first dimension
        Ny: amount of pixels to pad in second dimension
    """
    b = np.zeros((a.shape[0] + Nx, a.shape[1] + Ny) + a.shape[2:]).astype(a.dtype)
    b[: a.shape[0], : a.shape[1]] = a
    return b



""".. _IOLib:

A library to handle reading/writing files.
"""

from __future__ import annotations

from pyloggrid.Libs.singlethread_numpy import np

np.zeros(0)  # Do not remove - used to force singlethread_numpy to be imported before the other libraries when refactoring

import logging
import os
import pathlib
import shutil
import typing
from typing import Callable, List

import h5py as h5py
import orjson

from pyloggrid.Libs.custom_logger import setup_custom_logger
from pyloggrid.LogGrid.Grid import Grid

if typing.TYPE_CHECKING:
    from pyloggrid.Libs.misc import TimeTracker

logger = setup_custom_logger(__name__, level=logging.INFO)


# Decorators
def store_hdf5(fun: Callable) -> Callable:
    """Decorator to handle fail-safe saving to hdf5.

    Attempts to save to ``<path>/output.h5``, then to ``<path>/output_bk.h5``.

    Note:
        The decorated function's path argument becomes a directory instead of a file.

    Args:
        fun: a non failsafe saving function that takes a ``path`` (hdf5 file) first argument

    Example:
        ::

            @store_hdf5
            def add_group(path: str, group_name: str) -> None:
                path: h5py.File  # from the decorator
                path.create_group(group_name)
    """

    # noinspection PyMissingOrEmptyDocstring
    def wrapper(path: str | pathlib.Path, *args, **kwargs) -> None:
        path = pathlib.Path(path)
        normal_path = path / "output.h5"
        bk_path = path / "output_bk.h5"
        if normal_path.exists() and not bk_path.exists():
            shutil.copy(normal_path, bk_path)
        with h5py.File(normal_path, "a") as f:
            fun(f, *args, **kwargs)
        with h5py.File(bk_path, "a") as f:
            fun(f, *args, **kwargs)

    return wrapper


def load_hdf5(fun: Callable) -> Callable:
    """
    Decorator to handle fail-safe loading from hdf5.

    Attempts to load from ``<path>/output.h5``, or to ``<path>/output_bk.h5`` if the first one fails.

    Note:
        The decorated function's path argument becomes a directory instead of a file.

    Warning:
        If the decorated ``path`` argument is a file instead of a directory, this will only load this file (non fail-safe).

    Args:
        fun: a non failsafe loading function that takes a ``path`` (hdf5 file) first argument

    Example:
        ::

            @load_hdf5
            def load_group(f: h5py.File, group_name: str) -> dict:
                group = f[group_name]
                return {name: group[name][:] for name in group}
    """

    # noinspection PyMissingOrEmptyDocstring
    def wrapper(path, filepath=None, *args, **kwargs):
        if filepath is not None:
            with h5py.File(filepath, "r") as f:
                return fun(f, *args, **kwargs)

        path = pathlib.Path(path)
        try:
            with h5py.File(path / "output.h5", "r") as f:
                return fun(f, *args, **kwargs)
        except OSError:  # file is corrupted / missing
            with h5py.File(path / "output_bk.h5", "r") as f:
                return fun(f, *args, **kwargs)

    return wrapper


# Json
def read_json_settings(path: str) -> dict:
    """reads the ``settings.json`` simulation file (fail-safe), returns the data

    Args:
        path: directory where to find the file

    Returns:
        the json data as a dict

    Raises:
        AssertionError: if the path doesn't exist
    """

    sett_path = f"{path}/settings.json"
    sett_bk_path = f"{path}/settings.json_bk"
    assert os.path.isfile(sett_path) or os.path.isfile(sett_bk_path), "The path to load data from doesn't exist"

    logging.debug(f"Attempting to read JSON settings from {sett_path}")

    try:
        with open(sett_path, "r") as f:
            settings = orjson.loads(f.read())
    except Exception as e:
        logger.warning(f"Could not read settings file ({e}), falling back on {sett_bk_path}")
        with open(sett_bk_path, "r") as f:
            settings = orjson.loads(f.read())

    return settings


def update_json_settings(path: str, params: dict, update: bool = False) -> None:
    """Save simulation parameters to ``settings.json`` file.

    This is fail-safe though the use of ``settings_bk.json``.

    Args:
        path: directory to save the settings to
        params: parameters to save to file
        update: if True, then existing parameters will be updated according to existing keys in ``params``. Otherwise, settings are overwritten.
    """
    params = dict(params)  # copy
    sett_path = f"{path}/settings.json"
    sett_bk_path = f"{path}/settings.json_bk"
    sett_new_path = f"{path}/settings.json_new"

    if update:  # update existing parameters with keys in `params`
        with open(sett_path, "r") as f:
            data = orjson.loads(f.read())
        for k, v in params.items():
            data[k] = v
        params = data

    if os.path.isfile(sett_path):
        shutil.copy(sett_path, sett_bk_path)  # backup existing file in case the data may get corrupted
    with open(sett_new_path, "wb") as f:
        f.write(orjson.dumps(params, option=orjson.OPT_NON_STR_KEYS | orjson.OPT_SERIALIZE_NUMPY))
    # only rename once the data has been written
    os.replace(sett_new_path, sett_path)


# Hdf5
def load_step(fields_names: List[str], path: str = None, step: int = None, filepath: str = None) -> dict:
    """Load the data corresponding to a simulation step saved in ``output.h5``, or older legacy formats.

    Only the ``path`` and ``filepath`` arguments are necessary in the latest version.

    Args:
        filepath: path to save file. ([legacy] If omitted, the combination of `path` and `step` will be used instead.)
        fields_names: [legacy] grid's fields' names.
        path: data's parent directory
        step: [legacy] step number to load

    Warning:
        The legacy support (ordered and unordered ``.npz``) may be deprecated in the future, and only exists for loading old results.

    Todo:
        Once we deprecate the legacy support (3.x ?), remove the additional obsolete arguments, and reflect in usages

    Returns:
        a dict containing the saved data, ``{fields: the grid's fields, t: step time, N_points: grid size, k_min: grid k min, elapsed_time: real elapsed time, k0: grid.k0}``
    """

    @load_hdf5
    def load_step_hdf5(f: h5py.File):
        """load for hdf5"""
        group = f[f"/step_{step}"]
        fields = {name: group[name][:] for name in group}
        step_data = {**dict(group.attrs), "fields": fields}
        return step_data

    try:  # hdf5 save
        return load_step_hdf5(path, filepath)
    except FileNotFoundError:  # Legacy support: .npz saves
        if filepath is None:
            filepath = f"{path}/fields_{step}.npz"
        data = np.load(filepath, allow_pickle=True)

        try:
            fields = {name: data[f"field_{name}"] for name in fields_names}
            step_data = {
                "fields": fields,
                "t": data["t"],
                "N_points": data["N_points"],
                "k_min": data["k_min"],
                "elapsed_time": data["elapsed_time"],
                "k0": data["k0"],
            }
        except KeyError:  # Legacy support for old npz save /!\ the order of `fields_names` is important
            reserved_arrays = 5
            fields = {fields_names[i]: data[f"arr_{reserved_arrays + i}"] for i in range(len(fields_names))}
            step_data = {
                "fields": fields,
                "t": data["arr_0"],
                "N_points": data["arr_1"],
                "k_min": data["arr_2"],
                "elapsed_time": data["arr_3"],
                "k0": data["arr_4"],
            }

        return step_data


@store_hdf5
def save_step(path: str, step: int, grid: Grid, t: float, dt: float, time_tracker: TimeTracker) -> None:
    """Save current step's fields to a .h5 file ``save.h5`` (appending if existing), as a new dataset ``step_<step>``

    Args:
        path: folder to save to
        grid: grid containing the fields to save
        t: time of the step
        dt: timestep used
        step: # of the step
        time_tracker: object that tracks elapsed time
    """
    path: h5py.File  # from the decorator
    group_name = f"step_{step}"
    group = path.create_group(group_name)
    group.attrs.create("t", t)
    group.attrs.create("N_points", grid.N_points)
    group.attrs.create("k_min", grid.k_min)
    group.attrs.create("k0", grid.k0)
    group.attrs.create("dt", dt)
    for k, v in time_tracker.elapsed_time.items():
        group.attrs.create(f"ttrack_{k}", v)
    for name in grid.fields_names:
        group.create_dataset(name, data=grid.fields[name])


# DataExplorer drawables
def add_drawables(path: str, drawables: dict) -> None:
    """Adds a drawables data dict to a DataExplorer result

    Args:
        path: the path where the DataExplorer result is
        drawables: the dict to save
    """
    # noinspection PyTypeChecker
    np.save(f"{path}/drawables", {"drawables": drawables})  # saved as an np.ndarray



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyloggrid', 'pyloggrid.Libs', 'pyloggrid.LogGrid']

package_data = \
{'': ['*']}

install_requires = \
['cython>=3.0.4,<4.0.0',
 'h5py>=3.8.0,<4.0.0',
 'imageio-ffmpeg>=0.4.8,<0.5.0',
 'imageio>=2.28.1,<3.0.0',
 'joblib>=1.2.0,<2.0.0',
 'matplotlib>=3.8.0,<4.0.0',
 'numpy>=1.26.1,<2.0.0',
 'orjson>=3.8.11,<4.0.0',
 'pyqt6>=6.5.3,<7.0.0',
 'rkstiff>=0.3.0,<0.4.0',
 'scienceplots>=2.0.1,<3.0.0',
 'scipy>=1.11.2,<2.0.0']

setup_kwargs = {
    'name': 'pyloggrid',
    'version': '2.3.1',
    'description': 'A python library to perform simulations on logarithmic lattices',
    'long_description': 'PyLogGrid is a Python-based framework for running and analyzing numerical simulations on [log-lattices [1]](https://www.doi.org/10.1088/1361-6544/abef73). The log-lattice structure is particularly useful for modeling phenomena that exhibit multi-scale behavior, such as turbulence. PyLogGrid is designed to be flexible, customizable, and easy to use.\n\nThis framework has been used in several scientific papers such as [[2]](https://www.doi.org/10.1017/jfm.2023.204), [[3]](https://www.doi.org/10.1103/PhysRevE.107.065106).\n\nThe framework includes a variety of built-in tools for analyzing simulation results, including visualization tools and post-processing scripts.\n\n### References:\n\n**[1]** Campolina, C. S., & Mailybaev, A. A. (2021). Fluid dynamics on logarithmic lattices. Nonlinearity, 34(7), 4684. doi:10.1088/1361-6544/abef73\n\n**[2]** Barral, A., & Dubrulle, B. (2023). Asymptotic ultimate regime of homogeneous Rayleigh–Bénard convection on logarithmic lattices. Journal of Fluid Mechanics, 962, A2. doi:10.1017/jfm.2023.204\n\n**[3]** Costa, G., Barral, A., & Dubrulle, B. (2023). Reversible Navier-Stokes equation on logarithmic lattices. Physical Review E, 107(6), 065106. doi:10.1103/PhysRevE.107.065106\n',
    'author': 'Amaury Barral',
    'author_email': 'amaury.barral@protonmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://drf-gitlab.cea.fr/amaury.barral/log-grid',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.12',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)
